unit ASyncSocket;

(*
 Asynchronous TCP Sokect Component
 (ASyncSocket is Delphi's socket component based on Winsock2)

 ASyncSocket.pas
   ASyncTCPClient : TCP Client
   ASyncTCPServer : TCP Server

   - Winsock2 based
   - IPv6 Support
   - Asynchronous TCP Server and Client
   - UnicodeString support (Delphi 2009 or abobe)
   - TextBase send/receive interface is available (WriteString() / ReadString() function)

   Version : 1.0.4
   Date : December 1, 2008
   Requirement : Delphi2009 (or abobe) and Winsock2 API (IdWinsock2.pas)
   Copyright : Copyright (C) 2008 XSEED Network

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

   ChangeLog
     1.0.4 (December 1, 2008)
       Append method
        - Reset() function has been implemented.
     1.0.3 (November 27, 2008)
       Bug fix
        - Component will not distab shutdown operating system.
     1.0.2 (November 26, 2008)
       Properties
        - OnWrite event has been changed (Wrote to the socket --> Ready to write to the socket)
        - ASyncTCPServer.ActiveConnections property has been added.
     1.0.1 (November 21, 2008)
       Client source address
        - BindAddress property has been implemented (Client source ip address).
       Bug fix
        - call freeaddrinfo() to release memory of old address information.
        - The Problem that cannot make new connection if call Disconnect() while connecting has been fixed.
     1.0.0 (November 20, 2008)
       First public release
       Protocol Version
        - Property of select protocol version is available.
     0.1.3 (November 19, 2008)
       Bug fix
        - Memory problem when free the ASyncTCPClient in resolving host has been fixed.
        - ASyncTCPServer.OnConnect event will invoke immediately after accept connection.
     0,1,2 (November 17, 2008)
       Super class change to TComponent
        - ASyncTCPClient and ASyncTCPServer can handle as VCL component.
        - Populer properties of these components can set on the property editor.
       Enhancement error processing
        - The error event has been able to handle when lookup fail.
        - Default error handling procedure show detail message.
        - Socket will be rightly destroyed when connection aborted. (bug fix)
        - All of error class in this component inherit from ESocketException
       Others
        - Call the API shutdown(how=1) before closesocket() when called Disconnect().
        - The control will not be blocked when connect in ASyncTCPClient
          - The previous version was blocked in getaddrinfo() API.
        - All comment of source has been translated to English.
     0,1,1 (November 15, 2008)
       First closed release.
*)

{$M+}

interface

uses
  Windows, Messages, Classes, SysUtils,
  Generics.Collections, RTLConsts,
  IdWinSock2;

type PPADDRINFO = ^PADDRINFO;
function getaddrinfo(NodeName: PAnsiChar; ServiceName: PAnsiChar; Hints: Paddrinfo; ppResult: PPADDRINFO): Integer; stdcall;
procedure freeaddrinfo(ai : PADDRINFO) ; stdcall;

const
  WM_SOCKETMESSAGE = WM_USER + $0001;
  CM_LOOKUPCOMPLEAT = WM_USER + $0002;

type
  TErrorEvent = (eeGeneral, eeSend, eeReceive, eeConnect, eeDisconnect, eeAccept, eeLookup);
  TSocketNotifyEvent = procedure(Sender : TObject ; Socket : TSocket) of object;
  TSocketNotifyErrorEvent = procedure(Sender : TObject ; Socket : TSocket ; ErrorEvent : TErrorEvent ; var ErrorCode : Integer) of object;
  TSocketNotifyTryEvent = procedure(Sender : TObject ; DestinationHost : UnicodeString ; DestinationPort : Integer) of object;
  TSocketNotifyAcceptEvent = procedure(Sender : TObject ; var CanAccept : boolean) of object;

  TCMSocketMessage = record
    Msg: Cardinal;
    Socket: TSocket;
    SelectEvent: Word;
    SelectError: Word;
    Result: Longint;
  end;

  TProtocolVersion = (IPAny, IPv4, IPv6);

  TAsyncBaseSocket = class(TComponent)
  protected
    m_hwnd : THandle;
    m_active : boolean;
    m_pver : TProtocolVersion;
    f_OnError : TSocketNotifyErrorEvent;
    f_OnConnect : TSocketNotifyEvent;
    f_OnDisconnect : TSocketNotifyEvent;
    f_OnAccept : TSocketNotifyAcceptEvent;
    f_OnRead : TSocketNotifyEvent;
    f_OnWrite : TSocketNotifyEvent;
    procedure Loaded; override;
    procedure WndProc(var Message: TMessage);
    procedure WMQueryEndSession(var Message: TWMQueryEndSession); message WM_QUERYENDSESSION;
    procedure WMSocketMessage(var Message: TCMSocketMessage); message WM_SOCKETMESSAGE;
    procedure EventError(Socket : TSocket ; ErrorEvent : TErrorEvent ; var ErrorCode : Integer); virtual;
    procedure EventConnect(Socket : TSocket); virtual;
    procedure EventDisconnect(Socket : TSocket); virtual;
    procedure EventAccept(ServerSocket : TSocket ; var ClientSocket : TSocket ; var CanAccept : boolean); virtual;
    procedure EventRead(Socket : TSocket); virtual;
    procedure EventWrite(Socket : TSocket); virtual;
    procedure SetActive(value : boolean); virtual; abstract;
    procedure FreeSocket(Socket : TSocket); virtual;
    property OnError : TSocketNotifyErrorEvent read f_OnError write f_OnError;
    property OnConnect : TSocketNotifyEvent read f_OnConnect write f_OnConnect;
    property OnDisconnect : TSocketNotifyEvent read f_OnDisconnect write f_OnDisconnect;
    property OnAccept : TSocketNotifyAcceptEvent read f_OnAccept write f_OnAccept;
    property OnRead : TSocketNotifyEvent read f_OnRead write f_OnRead;
    property OnWrite : TSocketNotifyEvent read f_OnWrite write f_OnWrite;
    property Active : boolean read m_active write SetActive default False;
    property ProtocolVersion : TProtocolVersion read m_pver write m_pver default IPAny;
  public
    constructor Create(Owner : TComponent); override;
    destructor Destroy(); override;
    function GetRemoteAddress(Socket : TSocket) : UnicodeString;
    function GetRemotePort(Socket : TSocket) : integer;
    function GetRemoteHost(Socket : TSocket) : UnicodeString;
    function Read(Socket : TSocket ; var buffer ; length : integer) : integer;
    function Write(Socket : TSocket ; const buffer ; length : integer) : integer;
    function ReadString(Socket : TSocket) : RawByteString;
    function WriteString(Socket : TSocket ; Text : RawByteString) : integer;
    function ReadUnicodeString(Socket : TSocket) : UnicodeString;
    function WriteUnicodeString(Socket : TSocket ; Text : UnicodeString) : integer;
    function Disconnect(Socket : TSocket) : integer;
    function Reset(Socket : TSocket) : integer;
  end;

  TTryConnect = record
    ai0 : PADDRINFO;
    aicurrent : PADDRINFO;
    thread : TThread;
  end;

  TASyncTCPClient = class(TAsyncBaseSocket)
  private
    m_socket : TSocket;
    m_tryconnect : TTryConnect;
    m_serverport : Integer;
    m_serverhost : UnicodeString;
    m_bindsrc : UnicodeString;
    f_tryconnect : TSocketNotifyTryEvent;
  protected
    procedure LookupMessage(var Message: TMessage); message CM_LOOKUPCOMPLEAT;
    procedure EventError(Socket : TSocket ; ErrorEvent : TErrorEvent ; var ErrorCode : Integer); override;
    procedure Activate();
    procedure SetActive(value : boolean); override;
    procedure FreeSocket(Socket : TSocket); override;
  public
    constructor Create(Owner : TComponent); override;
    destructor Destroy(); override;
    function GetRemoteAddress() : UnicodeString;
    function GetRemotePort() : integer;
    function GetRemoteHost() : UnicodeString;
    function Read(var buffer ; length : integer) : integer;
    function Write(const buffer ; length : integer) : integer;
    function ReadString() : RawByteString;
    function WriteString(Text : RawByteString) : integer;
    function ReadUnicodeString() : UnicodeString;
    function WriteUnicodeString(Text : UnicodeString) : integer;
    function Disconnect() : integer;
    function Reset() : integer;
    property ClientSocket : TSocket read m_socket;
    property BindAddress : UnicodeString read m_bindsrc write m_bindsrc;
  published
    property Port : Integer read m_serverport write m_serverport default 0;
    property Host : UnicodeString read m_serverhost write m_serverhost;
    property OnError;
    property OnConnect;
    property OnDisconnect;
    property OnRead;
    property OnWrite;
    property Active;
    property ProtocolVersion;
    property OnTryToConnect : TSocketNotifyTryEvent read f_tryconnect write f_tryconnect;
  end;

  TListenItem = record
    ListenHost : UnicodeString;
    ListenPort : Integer;
  end;
  TListenList = TList<TListenItem>;
  TSocketList = TList<TSocket>;

  TASyncTCPServer = class(TAsyncBaseSocket)
  private
    m_list : TListenList;
    m_serversockets : TSocketList;
    m_clientsockets : TSocketList;
    m_defaultport : Integer;
  protected
    procedure EventAccept(ServerSocket : TSocket ; var ClientSocket : TSocket ; var CanAccept : boolean); override;
    procedure SetActive(value : boolean); override;
    procedure FreeSocket(Socket : TSocket); override;
    procedure SetDefaultBindPort(value : integer);
    function GetActiveConnections() : Integer;
  public
    constructor Create(Owner : TComponent); override;
    destructor Destroy(); override;
    procedure AppendBind(ListenPort : Integer);
    property ActiveConnections : Integer read GetActiveConnections;
    property ServerSockets : TSocketList read m_serversockets;
    property Connections : TSocketList read m_clientsockets;
    property Binds : TListenList read m_list;
  published
    property Port : Integer read m_defaultport write SetDefaultBindPort default 0;
    property OnError;
    property OnConnect;
    property OnDisconnect;
    property OnAccept;
    property OnRead;
    property OnWrite;
    property Active;
    property ProtocolVersion;
  end;

  TLookupThread = class(TThread)
  private
    m_hwnd : THandle;
    m_serverhost : UnicodeString;
    m_serverport : Integer;
    m_pver : TProtocolVersion;
  public
    constructor Create(hWnd: Cardinal ; host : UnicodeString ; port : Integer ; ver : TProtocolVersion); overload;
  protected
    procedure Execute; override;
  end;

  ESocketException = class(Exception)
  end;

  EASyncSocketException = class(ESocketException)
  public
    constructor Create(ErrorCode : Integer );
  end;

procedure Register;

implementation

//--------------------------------- APIs ---------------------------------------

procedure Register;
begin
  RegisterComponents('Internet', [TASyncTCPClient]);
  RegisterComponents('Internet', [TASyncTCPServer]);
end;

function getaddrinfo(NodeName: PAnsiChar; ServiceName: PAnsiChar; Hints: Paddrinfo; ppResult: PPADDRINFO): Integer; stdcall;
var
  proc : function(NodeName: PAnsiChar; ServiceName: PAnsiChar; Hints: Paddrinfo; ppResult: PPADDRINFO): Integer; stdcall;
  lib : THandle;
begin
  lib:=LoadLibrary('ws2_32.dll');
  if (lib=0) then
  begin
    result:=-1;
    exit;
  end;
  @proc:=GetProcAddress(lib,'getaddrinfo');
  if (@proc=nil) then
  begin
    result:=-1;
    exit;
  end;
  result:=proc(NodeName,ServiceName,Hints,ppResult);
  FreeLibrary(lib);
end;

procedure freeaddrinfo(ai : PADDRINFO); stdcall;
var
  proc : function(ai : PADDRINFO): Integer; stdcall;
  lib : THandle;
begin
  lib:=LoadLibrary('ws2_32.dll');
  if (lib=0) then
  begin
    exit;
  end;
  @proc:=GetProcAddress(lib,'freeaddrinfo');
  if (@proc=nil) then
  begin
    exit;
  end;
  FreeLibrary(lib);
end;

//--------------------------------- TAsyncBaseSocket ---------------------------------------

constructor TAsyncBaseSocket.Create(Owner : TComponent);
begin
  inherited Create(Owner);
  m_hwnd:=AllocateHWnd(WndProc);
end;

destructor TAsyncBaseSocket.Destroy();
begin
  DeallocateHWnd(m_hwnd);
  inherited;
end;

procedure TAsyncBaseSocket.Loaded;
begin
  inherited Loaded;
  if m_active then
  begin
    setActive(false);
    SetActive(true);
  end;
end;

procedure TAsyncBaseSocket.WndProc(var Message: TMessage);
begin
  try
    Dispatch(Message);
  except
    if Assigned(ApplicationHandleException) then
      ApplicationHandleException(Self);
  end;
end;

procedure TAsyncBaseSocket.WMQueryEndSession(var Message: TWMQueryEndSession);
begin
  Message.Result:=1;
end;

procedure TAsyncBaseSocket.WMSocketMessage(var Message: TCMSocketMessage);

  function CheckError: Boolean;
  var
    ErrorEvent: TErrorEvent;
    ErrorCode: Integer;
  begin
    if Message.SelectError <> 0 then
    begin
      Result := False;
      ErrorCode := Message.SelectError;
      case Message.SelectEvent of
        FD_CONNECT: ErrorEvent := eeConnect;
        FD_CLOSE: ErrorEvent := eeDisconnect;
        FD_READ: ErrorEvent := eeReceive;
        FD_WRITE: ErrorEvent := eeSend;
        FD_ACCEPT: ErrorEvent := eeAccept;
      else
        ErrorEvent := eeGeneral;
      end;
      EventError(Message.Socket, ErrorEvent, ErrorCode);
      if (ErrorCode<>0) then
       raise EASyncSocketException.Create(ErrorCode);
    end
    else Result := True;
  end;
var
  CanAccept : boolean;
  ClientSocket : TSocket;
begin
  CanAccept:=true;
  with Message do
    if CheckError then
      case SelectEvent of
        FD_CONNECT: EventConnect(Socket);
        FD_CLOSE: EventDisconnect(Socket);
        FD_READ: EventRead(Socket);
        FD_WRITE: EventWrite(Socket);
        FD_ACCEPT: EventAccept(Socket,ClientSocket,CanAccept);
      end;
end;

procedure TAsyncBaseSocket.EventError(Socket : TSocket ; ErrorEvent : TErrorEvent ; var ErrorCode : Integer);
begin
  if (@OnError<>nil) then
    OnError(self,Socket,ErrorEvent,ErrorCode);
  if ((ErrorEvent=eeDisconnect) or (ErrorEvent=eeConnect)) then
    FreeSocket(Socket);
end;

procedure TAsyncBaseSocket.EventConnect(Socket : TSocket);
begin
  if (@OnConnect<>nil) then
    OnConnect(self,Socket);
end;

procedure TAsyncBaseSocket.EventDisconnect(Socket : TSocket);
begin
  if (@OnDisconnect<>nil) then
    OnDisconnect(self,Socket);
  FreeSocket(Socket);
end;

procedure TAsyncBaseSocket.EventAccept(ServerSocket : TSocket ; var ClientSocket : TSocket ; var CanAccept : boolean);
var
  Addr: TSockAddrIn6;
  Len: Integer;
begin
  if (@OnAccept<>nil) then
    OnAccept(self,CanAccept);
  if CanAccept then
  begin
    Len:=sizeof(Addr);
    ClientSocket:=accept(ServerSocket,@Addr,@Len);
  end;
end;

procedure TAsyncBaseSocket.EventRead(Socket : TSocket);
begin
  if (@OnRead<>nil) then
    OnRead(self,Socket);
end;

procedure TAsyncBaseSocket.EventWrite(Socket : TSocket);
begin
  if (@OnWrite<>nil) then
    OnWrite(self,Socket);
end;

procedure TAsyncBaseSocket.FreeSocket(Socket : TSocket);
begin
  closesocket(Socket);
end;

function TAsyncBaseSocket.GetRemoteAddress(Socket : TSocket) : UnicodeString;
var
  in_addr : TSockAddrIn6;
  in_addr_size : integer;
  buf : array [0..1024] of WideChar;
  buf_size : cardinal;
begin
  FillChar(buf,sizeof(buf),0);
  in_addr_size:=sizeof(in_addr);
  if (getpeername(Socket,@in_addr,in_addr_size)<>ERROR_SUCCESS) then exit;
  buf_size:=sizeof(buf) div sizeof(WideChar);
  in_addr.sin6_port:=0; //do not need portno from WSAAddressToString
  WSAAddressToString(@in_addr,in_addr_size,nil,buf,buf_size);
  Result:=StrPas(buf);
end;

function TAsyncBaseSocket.GetRemotePort(Socket : TSocket) : integer;
var
  in_addr : TSockAddrIn6;
  in_addr_size : integer;
begin
  Result:=0;
  in_addr_size:=sizeof(in_addr);
  if (getpeername(Socket,@in_addr,in_addr_size)<>ERROR_SUCCESS) then exit;
  Result:=ntohs(in_addr.sin6_port);
end;

function TAsyncBaseSocket.GetRemoteHost(Socket : TSocket) : UnicodeString;
var
  in_addr : TSockAddrIn6;
  in_addr_size : integer;
  in_addr4 : TSockAddrIn;
  buf : array [0..1024] of WideChar;
  buf_size : cardinal;
  Ent : PHostEnt;
begin
  FillChar(buf,sizeof(buf),0);
  in_addr_size:=sizeof(in_addr);
  if (getpeername(Socket,@in_addr,in_addr_size)<>ERROR_SUCCESS) then exit;
  buf_size:=sizeof(buf) div sizeof(WideChar);
  WSAAddressToString(@in_addr,in_addr_size,nil,buf,buf_size);
  if (buf[0]='[') then
  begin
    Ent := gethostbyaddr(@in_addr.sin6_addr.s6_addr, 16, PF_INET6);
    if Ent <> nil then
      result:=UnicodeString(StrPas(Ent.h_name))
  end
  else
  begin
    in_addr_size:=sizeof(in_addr4);
    if (getpeername(Socket,@in_addr4,in_addr_size)<>ERROR_SUCCESS) then exit;
    Ent := gethostbyaddr(@in_addr4.sin_addr.s_addr, 4, PF_INET);
    if Ent <> nil then
      result:=UnicodeString(StrPas(Ent.h_name));
  end;
end;

function TAsyncBaseSocket.Read(Socket : TSocket ; var buffer ; length : integer) : integer;
begin
  result:=recv(Socket,buffer,length,0);
end;

function TAsyncBaseSocket.Write(Socket : TSocket ; const buffer ; length : integer) : integer;
begin
  result:=send(Socket,buffer,length,0);
end;

function TAsyncBaseSocket.ReadString(Socket : TSocket) : RawByteString;
var
  buffer : pansichar;
  length : integer;
begin
  ioctlsocket(Socket, FIONREAD, Cardinal(length));
  GetMem(buffer,length+1);
  FillChar(buffer^,length+1,0);
  Read(Socket,buffer^,length);
  result:=StrPas(buffer);
  FreeMem(buffer);
end;

function TAsyncBaseSocket.WriteString(Socket : TSocket ; Text : RawByteString) : integer;
begin
  result:=Write(Socket,pansichar(Text)^,StrLen(pansichar(Text)));
end;

function TAsyncBaseSocket.ReadUnicodeString(Socket : TSocket) : UnicodeString;
var
  buffer : pwidechar;
  length : integer;
begin
  ioctlsocket(Socket, FIONREAD, Cardinal(length));
  GetMem(buffer,length+2);
  FillChar(buffer^,length+2,0);
  Read(Socket,buffer^,length);
  result:=StrPas(buffer);
  FreeMem(buffer);
end;

function TAsyncBaseSocket.WriteUnicodeString(Socket : TSocket ; Text : UnicodeString) : integer;
begin
  result:=Write(Socket,pwidechar(Text)^,StrLen(pwidechar(Text))*2);
end;

function TAsyncBaseSocket.Disconnect(Socket : TSocket) : integer;
begin
  result:=shutdown(Socket,1);
end;

function TAsyncBaseSocket.Reset(Socket : TSocket) : integer;
var
  i : linger;
begin
  i.l_onoff:=1;
  i.l_linger:=0;
  result:=setsockopt(Socket,SOL_SOCKET,SO_LINGER,pAnsiChar(@i),sizeof(i));
  if (result=ERROR_SUCCESS) then
    FreeSocket(Socket);
end;

//--------------------------------- TASyncTCPClient ---------------------------------------

constructor TASyncTCPClient.Create(Owner : TComponent);
begin
  inherited Create(Owner);
  m_serverport:=0;
  m_serverhost:='';
  m_socket:=INVALID_SOCKET;
  m_tryconnect.ai0:=nil;
end;

destructor TASyncTCPClient.Destroy;
begin
  //wait for terminate current lookup thread
  if Assigned(m_tryconnect.thread) then
  begin
    m_tryconnect.thread.FreeOnTerminate:=false;
    m_tryconnect.thread.Free;
  end;
  if (m_tryconnect.ai0<>nil) then
    freeaddrinfo(m_tryconnect.ai0);
  SetActive(false);
  inherited;
end;


procedure TASyncTCPClient.LookupMessage(var Message: TMessage);
var
  ErrorCode : Integer;
begin
  m_tryconnect.thread:=nil;
  if (m_tryconnect.ai0<>nil) then //Old information
    freeaddrinfo(m_tryconnect.ai0);

  if (Message.WParam = ERROR_SUCCESS) then
  begin
    m_tryconnect.ai0 := Pointer(Message.LParam);
    m_tryconnect.aicurrent := m_tryconnect.ai0;
    Activate();
  end;
  if (Message.WParam = SOCKET_ERROR) then
  begin
    ErrorCode:=Message.LParam;
    EventError(INVALID_SOCKET, eeLookup, ErrorCode);
    setActive(false);
    if (ErrorCode<>0) then
      raise ESocketException.CreateResFmt(@sWindowsSocketError, [SysErrorMessage(ErrorCode), ErrorCode, 'gethostbyname']);
  end;
end;

procedure TASyncTCPClient.EventError(Socket : TSocket ; ErrorEvent : TErrorEvent ; var ErrorCode : Integer);
begin
  if (ErrorEvent=eeConnect) then
  begin
    closesocket(m_socket);
    m_socket := INVALID_SOCKET;
  end;

  if ( ((ErrorEvent=eeConnect) or (ErrorEvent=eeLookup)) and not m_active) then
  begin
    //abort connection
    ErrorCode:=0;
  end
  else if ((ErrorEvent=eeConnect) and (m_tryconnect.aicurrent<>nil)) then
  begin
    //try next connection hint
    ErrorCode:=0;
    Activate();
  end
  else
    inherited EventError(Socket,ErrorEvent,ErrorCode);
end;

procedure TASyncTCPClient.Activate();
var
  ai : LPADDRINFO;
  sock : TSocket;
  buf : array[0..1024] of WideChar;
  buf_size : Cardinal;
  port : Word;
  srcai : LPADDRINFO;
  srchints : ADDRINFO;
  ErrorCode : integer;
begin
  srcai:=nil;
  if (m_bindsrc<>'') then
  begin
    FillChar(srchints,sizeof(srchints),0);
    srchints.ai_family := AF_UNSPEC;
    srchints.ai_socktype := SOCK_STREAM;
    srchints.ai_flags :=  $0000000C;//AI_NUMERICHOST+AI_NUMERICSERV
    if (getaddrinfo(pAnsiChar(RawByteString(m_bindsrc)), nil, @srchints, @srcai)<>ERROR_SUCCESS) then
      srcai:=nil;
  end;

  ai := m_tryconnect.aicurrent;
  while((ai <> nil) and (m_socket = INVALID_SOCKET)) do
  begin
    buf_size:=sizeof(buf) div sizeof(WideChar);
    port:=ai.ai_addr.sin_port;
    ai.ai_addr.sin_port:=0;
    WSAAddressToString(ai.ai_addr,ai.ai_addrlen,nil,buf,buf_size);
    ai.ai_addr.sin_port:=port;
    if (@OnTryToConnect<>nil) then
      OnTryToConnect(self,StrPas(buf),ntohs(port));

    sock:=socket(ai.ai_family, ai.ai_socktype, ai.ai_protocol);
    if (sock = INVALID_SOCKET) then
    begin
      ai := ai.ai_next;
      continue;
    end;

    if (srcai<>nil) then
    begin
      if (bind(sock,srcai.ai_addr,srcai.ai_addrlen) = SOCKET_ERROR) then
      begin
        closesocket(sock);
        ai := ai.ai_next;
        continue;
      end;
    end;

    if (WSAAsyncSelect(sock, m_hwnd, WM_SOCKETMESSAGE, FD_CONNECT or FD_READ or FD_WRITE or FD_CLOSE) = SOCKET_ERROR) then
    begin
      closesocket(sock);
      ai := ai.ai_next;
      continue;
    end;

    connect(sock, ai.ai_addr, ai.ai_addrlen); //Always fail (because ASync Mode)

    m_socket:=sock;
    m_tryconnect.aicurrent:=ai.ai_next;
    m_active:=true;
  end;
  if (m_socket=INVALID_SOCKET) then
  begin
    ErrorCode:=10049;
    m_tryconnect.aicurrent:=nil;
    EventError(INVALID_SOCKET,eeConnect,ErrorCode);
    if (ErrorCode<>0) then
      raise ESocketException.CreateResFmt(@sAsyncSocketError, [ErrorCode]);
  end;
end;

procedure TASyncTCPClient.SetActive(value : boolean);
begin
  if (csDesigning in ComponentState) or (csLoading in ComponentState) then
  begin
    m_active := value;
    exit;
  end;

  if value and not m_active then
  begin
    m_tryconnect.thread:=TLookupThread.Create(m_hwnd,m_serverhost,m_serverport,m_pver);
    m_tryconnect.thread.FreeOnTerminate:=true;
    m_tryconnect.thread.Suspended:=false;
    m_active:=true;
  end
  else if m_active then
  begin
    if m_tryconnect.thread<>nil then
    begin
      m_tryconnect.thread.Terminate;
      m_tryconnect.thread:=nil;
    end;
    Disconnect();
    m_active:=false;
  end;
end;

procedure TASyncTCPClient.FreeSocket(Socket : TSocket);
begin
  m_socket:=INVALID_SOCKET;
  m_active:=false;
  inherited FreeSocket(Socket);
end;

function TASyncTCPClient.GetRemoteAddress() : UnicodeString;
begin
  result := inherited GetRemoteAddress(m_socket);
end;
function TASyncTCPClient.GetRemotePort() : integer;
begin
  result := inherited GetRemotePort(m_socket);
end;
function TASyncTCPClient.GetRemoteHost() : UnicodeString;
begin
  result := inherited GetRemoteHost(m_socket);
end;

function TASyncTCPClient.Read(var buffer ; length : integer) : integer;
begin
  result := inherited Read(m_socket,buffer,length);
end;

function TASyncTCPClient.Write(const buffer ; length : integer) : integer;
begin
  result := inherited Write(m_socket,buffer,length);
end;

function TASyncTCPClient.ReadString() : RawByteString;
begin
  result := inherited ReadString(m_socket);
end;

function TASyncTCPClient.WriteString(Text : RawByteString) : integer;
begin
  result := inherited WriteString(m_socket,Text);
end;

function TASyncTCPClient.ReadUnicodeString() : UnicodeString;
begin
  result := inherited ReadUnicodeString(m_socket);
end;

function TASyncTCPClient.WriteUnicodeString(Text : UnicodeString) : integer;
begin
  result := inherited WriteUnicodeString(m_socket,Text);
end;

function TASyncTCPClient.Disconnect() : integer;
begin
  result := inherited Disconnect(m_socket);
end;

function TASyncTCPClient.Reset() : integer;
begin
  result := inherited Reset(m_socket);
end;

//--------------------------------- TASyncTCPServer ---------------------------------------

constructor TASyncTCPServer.Create(Owner : TComponent);
begin
  inherited Create(Owner);
  m_list:=TListenList.Create;
  m_clientsockets:=TSocketList.Create;
  m_serversockets:=TSocketList.Create;
end;

procedure TASyncTCPServer.AppendBind(ListenPort : Integer);
var
  item : TListenItem;
begin
  item.ListenHost:=''; //Any Address
  item.ListenPort:=ListenPort;
  m_list.Add(item);
end;

destructor TASyncTCPServer.Destroy();
begin
  SetActive(false);
  m_list.Free;
  m_clientsockets.Free;
  m_serversockets.Free;
  inherited;
end;

procedure TASyncTCPServer.EventAccept(ServerSocket : TSocket ; var ClientSocket : TSocket ; var CanAccept : boolean);
begin
  inherited EventAccept(ServerSocket,ClientSocket,CanAccept);
  if (not CanAccept) then exit;
  m_clientsockets.Add(ClientSocket);
  EventConnect(ClientSocket);
end;

procedure TASyncTCPServer.SetActive(value : boolean);
var
  host : pAnsiChar;
  hints : ADDRINFO;
  ai, ai0 : LPADDRINFO;
  ret,nsocks,nlisten : integer;
  sock : TSocket;
begin
  if (csDesigning in ComponentState) or (csLoading in ComponentState) then
  begin
    m_active := value;
    exit;
  end;

  if value and not m_active then
  begin
    FillChar(hints,sizeof(hints),0);
    hints.ai_family := AF_UNSPEC;
		hints.ai_socktype := SOCK_STREAM;
		hints.ai_flags :=  $0000000D; //AI_PASSIVE+AI_NUMERICHOST+AI_NUMERICSERV
    if (m_pver=IPv4) then hints.ai_family := AF_INET;
    if (m_pver=IPv6) then hints.ai_family := AF_INET6;

    nsocks := 0;

    for nlisten := 0 to m_list.Count - 1 do
    begin
      host:=nil;
      if (m_list[nlisten].ListenHost<>'') then
        host:=pAnsiChar(RawByteString(m_list[nlisten].ListenHost));
      ret:=getaddrinfo(host, pAnsiChar(RawByteString(IntToStr(m_list[nlisten].ListenPort))), @hints, @ai0);
      if (ret<>ERROR_SUCCESS) then continue;

      ai := ai0;
      while(ai <> nil) do
      begin
        sock:=socket(ai.ai_family, ai.ai_socktype, ai.ai_protocol);
        if (sock = INVALID_SOCKET) then
        begin
          ai := ai.ai_next;
          continue;
        end;

        if (WSAAsyncSelect(sock, m_hwnd, WM_SOCKETMESSAGE, FD_ACCEPT or FD_READ or FD_WRITE or FD_CLOSE) = SOCKET_ERROR) then
        begin
          closesocket(sock);
          ai := ai.ai_next;
          continue;
        end;

        if (bind(sock, ai.ai_addr, ai.ai_addrlen) = SOCKET_ERROR) then
        begin
          closesocket(sock);
          ai := ai.ai_next;
          continue;
        end;

        if (listen(sock, 5) = SOCKET_ERROR) then
        begin
			  	closesocket(sock);
          ai := ai.ai_next;
          continue;
        end;

        m_serversockets.Add(sock);
        nsocks := nsocks+1;
        ai := ai.ai_next;
      end;
    end;
    freeaddrinfo(ai0);
    if (nsocks = 0) then
      raise ESocketException.CreateResFmt(@sCannotListenOnOpen,[])
    else
      m_active:=true;
  end
  else if m_active then
  begin
    //Disable new connection
    for nsocks := 0 to m_serversockets.Count - 1 do
    begin
      closesocket(m_serversockets[nsocks]);
    end;
    m_serversockets.Clear;
    //Disable current connection
    for nsocks := 0 to m_clientsockets.Count - 1 do
    begin
      shutdown(m_clientsockets[nsocks],2);
      closesocket(m_clientsockets[nsocks]);
    end;
    m_clientsockets.Clear;
    m_active:=false;
  end;
end;

procedure TASyncTCPServer.FreeSocket(Socket : TSocket);
var
  i,index : integer;
begin
  index:=-1;
  for i := 0 to m_clientsockets.Count - 1 do
    if (m_clientsockets[i]=Socket) then
      index:=i;
  if (index>=0) then
    m_clientsockets.Delete(index);

  inherited FreeSocket(Socket);
end;

procedure TASyncTCPServer.SetDefaultBindPort(value : integer);
begin
  m_defaultport:=value;
  if (value>0) then
  begin
    m_list.Clear;
    AppendBind(value);
  end;
end;

function TASyncTCPServer.GetActiveConnections() : Integer;
begin
  result:=m_clientsockets.Count;
end;

constructor TLookupThread.Create(hWnd: Cardinal ; host : UnicodeString ; port : Integer ; ver : TProtocolVersion);
begin
  Inherited Create(true);
  m_hwnd := hWnd;
  m_serverhost:=host;
  m_serverport:=port;
  m_pver:=ver;
end;

procedure TLookupThread.Execute;
var
  ret : integer;
  hints : ADDRINFO;
  ai0 : LPADDRINFO;
begin
  inherited;
  FillChar(hints,sizeof(hints),0);
  hints.ai_family := AF_UNSPEC;
  hints.ai_socktype := SOCK_STREAM;
  hints.ai_flags :=  $00000008; //AI_NUMERICSERV
  if (m_pver=IPv4) then hints.ai_family := AF_INET;
  if (m_pver=IPv6) then hints.ai_family := AF_INET6;

  ret:=getaddrinfo(pAnsiChar(RawByteString(m_serverhost)), pAnsiChar(RawByteString(IntToStr(m_serverport))), @hints, @ai0);
  if Terminated then
  begin
    freeaddrinfo(ai0);
    exit;
  end;
  if (ret<>ERROR_SUCCESS) then
    PostMessage(m_hwnd,CM_LOOKUPCOMPLEAT,SOCKET_ERROR,WSAGetLastError)
  else
    PostMessage(m_hwnd,CM_LOOKUPCOMPLEAT,ERROR_SUCCESS,Integer(ai0));
end;

constructor EASyncSocketException.Create(ErrorCode: Integer);
begin
  inherited CreateResFmt(@sASyncSocketError,[ErrorCode]);
end;

initialization
  InitializeWinSock; //WSAStartup

finalization
  UninitializeWinSock; //WSACleanup

end.

